/*
	ASCII MSX1 Emulator 'yaMSX1'
	Pioneer PX-7 Emulator 'ePX-7'
	ASCII MSX2 with DEMPA Sensor Kid Emulator 'yayaMSX2SK'

	Author : tanam
	Date   : 2013.06.29-

	modified by Takeda.Toshiya, umaiboux

	[ virtual machine ]
*/

#include "msx.h"
#include "../../emu.h"
#include "../device.h"
#include "../event.h"

#include "../datarec.h"
#include "../i8255.h"
#include "../io.h"
#include "../ld700.h"
#include "../not.h"
#include "../ym2203.h"
#include "../pcm1bit.h"
#include "../tms9918a.h"
#include "../v99x8.h"
#include "../z80.h"

#ifdef USE_DEBUGGER
#include "../debugger.h"
#endif

#include "joystick.h"
#include "keyboard.h"
#include "memory.h"

#include "../rp5c01.h"
#include "rtcif.h"
#include "sensorkid.h"

// ----------------------------------------------------------------------------
// initialize
// ----------------------------------------------------------------------------

VM::VM(EMU* parent_emu) : emu(parent_emu)
{
	// create devices
	first_device = last_device = NULL;
	dummy = new DEVICE(this, emu);	// must be 1st device
	event = new EVENT(this, emu);	// must be 2nd device
	
	drec = new DATAREC(this, emu);
	pio = new I8255(this, emu);
	io = new IO(this, emu);
#if defined(_PX7)
	ldp = new LD700(this, emu);
#endif
	not = new NOT(this, emu);
	psg = new YM2203(this, emu);
	pcm = new PCM1BIT(this, emu);
#if defined(_PX7)
	vdp = new TMS9918A(this, emu);
#elif defined(_MSX2SK)
	vdp = new V99X8(this, emu);
	rtc = new RP5C01(this, emu);
	rtcif = new RTCIF(this, emu);
	sensorkid = new SENSORKID(this, emu);
#endif
	cpu = new Z80(this, emu);
	
	joystick = new JOYSTICK(this, emu);
	keyboard = new KEYBOARD(this, emu);
	memory = new MEMORY(this, emu);
#if defined(_MSX2SK)
	slotmain = new SLOT_ROMMAIN(this, emu);
	cart[0] = new SLOT_CART(this, emu);
	cart[1] = new SLOT_CART(this, emu);
	slotsub = new SLOT_ROMSUB(this, emu);
	slotram = new SLOT_RAM64K(this, emu);
	slot_empty = new DEVICE(this, emu);
#else
	slot0 = new SLOT_ROMRAM32K(this, emu);
	cart[0] = new SLOT_CART(this, emu);
	slot2 = new SLOT_LDC(this, emu);
	cart[1] = new SLOT_CART(this, emu);
#endif
	
	// set contexts
	event->set_context_cpu(cpu);
	event->set_context_sound(psg);
	event->set_context_sound(pcm);
#if defined(_PX7)
	event->set_context_sound(ldp);
#endif
	
	drec->set_context_out(psg, SIG_YM2203_PORT_A, 0x80);
	pio->set_context_port_a(memory, SIG_MEMORY_SEL, 0xff, 0);
	pio->set_context_port_c(keyboard, SIG_KEYBOARD_COLUMN, 0x0f, 0);
#if defined(_PX7)
	pio->set_context_port_c(slot2, SIG_LDC_MUTE, 0x10, 0);
#endif
	pio->set_context_port_c(not, SIG_NOT_INPUT, 0x10, 0);
	not->set_context_out(drec, SIG_DATAREC_REMOTE, 1);
#if defined(_PX7)
	ldp->set_context_exv(slot2, SIG_LDC_EXV, 1);
	ldp->set_context_ack(slot2, SIG_LDC_ACK, 1);
	ldp->set_context_sound(psg, SIG_YM2203_PORT_A, 0x80);
#endif
	pio->set_context_port_c(drec, SIG_DATAREC_OUT, 0x20, 0);
	pio->set_context_port_c(pcm, SIG_PCM1BIT_SIGNAL, 0x80, 0);
	psg->set_context_port_b(joystick, SIG_JOYSTICK_SEL, 0x40, 0);
	vdp->set_context_irq(cpu, SIG_CPU_IRQ, 1);
	
	joystick->set_context_psg(psg);
	keyboard->set_context_cpu(cpu);
	keyboard->set_context_pio(pio);
#if defined(_MSX2SK)
	memory->set_context_slot_expand(false, false, false, true);
	memory->set_context_slot(0, 0, slotmain);
	memory->set_context_slot(1, 0, cart[0]);
	memory->set_context_slot(2, 0, cart[1]);
	memory->set_context_slot(3, 0, slotram);
	memory->set_context_slot(3, 1, slot_empty);
	memory->set_context_slot(3, 2, slot_empty);
	memory->set_context_slot(3, 3, slotsub);
	rtcif->set_context_rtc(rtc);
	io->set_iomap_range_rw(0xb4, 0xb5, rtcif);
	io->set_iomap_range_rw(0x00, 0x01, sensorkid);
#else
	memory->set_context_slot_expand(false, false, false, false);
	memory->set_context_slot(0, 0, slot0);
	memory->set_context_slot(1, 0, cart[0]);
	memory->set_context_slot(2, 0, slot2);
	memory->set_context_slot(3, 0, cart[1]);
#endif
#if defined(_PX7)
	slot2->set_context_cpu(cpu);
	slot2->set_context_ldp(ldp);
	slot2->set_context_vdp(vdp);
#endif
	
	// cpu bus
	cpu->set_context_mem(memory);
	cpu->set_context_io(io);
	cpu->set_context_intr(dummy);
#ifdef USE_DEBUGGER
	cpu->set_context_debugger(new DEBUGGER(this, emu));
#endif
	
	// i/o bus
#if defined(_MSX2SK)
	io->set_iomap_range_rw(0x98, 0x9b, vdp);
#else
	io->set_iomap_range_rw(0x98, 0x99, vdp);
#endif
	io->set_iomap_range_rw(0xa8, 0xab, pio);
	
	io->set_iomap_alias_w(0xa0, psg, 0);	// PSG ch
	io->set_iomap_alias_w(0xa1, psg, 1);	// PSG data
	io->set_iomap_alias_r(0xa2, psg, 1);	// PSG data
	
	// initialize all devices
	for(DEVICE* device = first_device; device; device = device->next_device) {
		device->initialize();
	}
}

VM::~VM()
{
	// delete all devices
	for(DEVICE* device = first_device; device;) {
		DEVICE *next_device = device->next_device;
		device->release();
		delete device;
		device = next_device;
	}
}

DEVICE* VM::get_device(int id)
{
	for(DEVICE* device = first_device; device; device = device->next_device) {
		if(device->this_device_id == id) {
			return device;
		}
	}
	return NULL;
}

// ----------------------------------------------------------------------------
// drive virtual machine
// ----------------------------------------------------------------------------

void VM::reset()
{
	// reset all devices
	for(DEVICE* device = first_device; device; device = device->next_device) {
		device->reset();
	}
}

void VM::run()
{
	event->drive();
}

// ----------------------------------------------------------------------------
// debugger
// ----------------------------------------------------------------------------

#ifdef USE_DEBUGGER
DEVICE *VM::get_cpu(int index)
{
	if(index == 0) {
		return cpu;
	}
	return NULL;
}
#endif

// ----------------------------------------------------------------------------
// draw screen
// ----------------------------------------------------------------------------

void VM::draw_screen()
{
	vdp->draw_screen();
}

// ----------------------------------------------------------------------------
// soud manager
// ----------------------------------------------------------------------------

void VM::initialize_sound(int rate, int samples)
{
	// init sound manager
	event->initialize_sound(rate, samples);
	
	// init sound gen
	psg->init(rate, 3579545, samples, 0, 0);
	pcm->init(rate, 8000);
#if defined(_PX7)
	ldp->initialize_sound(rate, samples);
#endif
}

uint16* VM::create_sound(int* extra_frames)
{
	return event->create_sound(extra_frames);
}

int VM::sound_buffer_ptr()
{
	return event->sound_buffer_ptr();
}

#if defined(_PX7)
void VM::movie_sound_callback(uint8 *buffer, long size)
{
	ldp->movie_sound_callback(buffer, size);
}
#endif

// ----------------------------------------------------------------------------
// user interface
// ----------------------------------------------------------------------------

void VM::open_cart(int drv, _TCHAR* file_path)
{
	if(drv < 2) {
		cart[drv]->open_cart(file_path);
	}
	reset();
}

void VM::close_cart(int drv)
{
	if(drv < 2) {
		cart[drv]->close_cart();
	}
	reset();
}

bool VM::cart_inserted(int drv)
{
	if(drv < 2) {
		return cart[drv]->cart_inserted();
	}
	return false;
}

void VM::play_tape(_TCHAR* file_path)
{
	drec->play_tape(file_path);
}

void VM::rec_tape(_TCHAR* file_path)
{
	drec->rec_tape(file_path);
}

void VM::close_tape()
{
	drec->close_tape();
}

bool VM::tape_inserted()
{
	return drec->tape_inserted();
}

#if defined(_PX7)
void VM::open_laser_disc(_TCHAR* file_path)
{
	ldp->open_disc(file_path);
}

void VM::close_laser_disc()
{
	ldp->close_disc();
}

bool VM::laser_disc_inserted()
{
	return ldp->disc_inserted();
}
#endif

bool VM::now_skip()
{
	return event->now_skip();
}

void VM::update_config()
{
	for(DEVICE* device = first_device; device; device = device->next_device) {
		device->update_config();
	}
}

